
/*
 *
 * br (BottleRocket)
 *
 * Control software for the X10(R) FireCracker wireless computer
 * interface kit.
 *
 * (c) 1999 Ashley Clark (aclark@ghoti.org) and Tymm Twillman (tymm@acm.org)
 *  Free Software.  LGPL applies.
 *  No warranties expressed or implied.
 *
 * Have fun with it and if you do anything really cool, send an email and let me
 * know.
 *
 */

#define VERSION "0.05b2"

#include "config.h"

#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <string.h>

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_FEATURES_H
#include <features.h>
#endif

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif

#include "br_cmd.h"
#include "br_cmd_engine.h"

#ifdef HAVE_ISSETUGID

/*
 * Thanks to Warner Losh for info on how to do this better
 */

#define ISSETID() (issetugid())
#else
#define ISSETID() (getuid() != geteuid() || getgid() != getegid())
#endif

#define SAFE_FILENO(fd) ((fd != STDIN_FILENO) && (fd != STDOUT_FILENO) \
                        && (fd != STDERR_FILENO))

int Verbose = 0;
char *MyName = "br";

void usage()
{
    fprintf(stderr, "BottleRocket version %s\n", VERSION);
    fprintf(stderr, "\n");
    fprintf(stderr, "Usage: %s [<options>][<housecode>(<list>) "
      "<native command> ...]\n\n", MyName);
    fprintf(stderr, "  Options:\n");
#ifdef HAVE_GETOPT_LONG
    fprintf(stderr, "  -v, --verbose\t\t\tadd v's to increase verbosity\n");
    fprintf(stderr, "  -x, --port=PORT\t\tset port to use\n");
    fprintf(stderr, "  -c, --house=[A-P]\t\tuse alternate house code "
      "(default \"A\")\n");
    fprintf(stderr, "  -n, --on=LIST\t\t\tturn on devices in LIST\n");
    fprintf(stderr, "  -f, --off=LIST\t\tturn off devices in LIST\n");
    fprintf(stderr, "  -N, --ON\t\t\tturn on all devices in housecode\n");
    fprintf(stderr, "  -F, --OFF\t\t\tturn off all devices in housecode\n");
    fprintf(stderr, "  -d, --dim=LEVEL[,LIST]\tdim devices in housecode to "
      " relative LEVEL\n");
    fprintf(stderr, "  -B, --lamps_on\t\tturn all lamps in housecode on\n");
    fprintf(stderr, "  -D, --lamps_off\t\tturn all lamps in housecode off\n");
    fprintf(stderr, "  -r, --repeat=NUM\t\trepeat commands NUM times "
      "(0 = ~ forever)\n");
    fprintf(stderr, "  -h, --help\t\t\tthis help\n\n");
#else
    fprintf(stderr, "  -v\tverbose (add v's to increase verbosity)\n");
    fprintf(stderr, "  -x\tset port to use\n");
    fprintf(stderr, "  -c\tuse alternate house code (default \"A\")\n");
    fprintf(stderr, "  -n\tturn on devices\n");
    fprintf(stderr, "  -f\tturn off devices\n");
    fprintf(stderr, "  -N\tturn all devices in housecode on\n");
    fprintf(stderr, "  -F\tturn all devices in housecode off\n");
    fprintf(stderr, "  -d\tdim devices in housecode to relative dimlevel\n");
    fprintf(stderr, "  -B\tturn all lamps in housecode on\n");
    fprintf(stderr, "  -D\tturn all lamps in housecode off\n");
    fprintf(stderr, "  -r\trepeat commands <repeats> times (0 basically "
      "means don't stop)\n");
    fprintf(stderr, "  -h\tthis help\n\n");
#endif
    fprintf(stderr, "<list>\t\tis a comma separated list of devices "
      "(no spaces),\n");
    fprintf(stderr, "\t\teach ranging from 1 to 16\n");     
    fprintf(stderr, "<dimlevel>\tis an integer from -6 to 6 "
      "(0 means no change)\n");
    fprintf(stderr, "<housecode>\tis a letter between A and P\n");
    fprintf(stderr, "<native cmd>\tis one of ON, OFF, DIM, BRIGHT, "
      "ALL_ON, ALL_OFF,\n");
    fprintf(stderr, "\t\tLAMPS_ON or LAMPS_OFF\n\n");
    fprintf(stderr, "For native commands, <list> should only be specified "
      "for ON or OFF.\n\n");

}

int checkimmutableport(char *port_source)
{
/*
 * Check to see if the user is allowed to specify an alternate serial port
 */

    if (!ISSETID())
        return 0;

    fprintf(stderr, "%s: You are not authorized to change the X10 port!\n",
      MyName);
    fprintf(stderr, "%s: Invalid port assignment %s.\n", MyName, port_source);

    errno = EINVAL;

    return -1;
}

int add_dimcmd(br_control_info *cinfo, int house, br_device_list *devs, int dim_level)
{
/*
 * Turn info into stuff usable by the command engine and add it to the
 *  list of things to do.
 */

    int index = 0;
    int dev;
    int cmd;

    cmd = ((dim_level < 0) ? DIM:BRIGHT);
    dim_level = (dim_level < 0) ? -dim_level:dim_level;

    if (br_get_num_devices(devs)) {
        while ((dev = br_get_next_device(devs, &index)) != -1) {
            if (br_add_cmd(cinfo, ON, house, dev, 0) < 0)
                return -1;
            if (br_add_cmd(cinfo, cmd, house, 0, dim_level) < 0)
                return -1;
        }
    } else {
        if (br_add_cmd(cinfo, cmd, house, 0, dim_level) < 0)
            return -1;
    }

    return 0;
}

int gethouse(char *house)
{
/*
 * Grab a house code from the command line
 */

    char *end;
    int c;


    c = br_strtohc(house, &end);

    if ((c  < 0) || (*end != '\0')) {
        fprintf(stderr, "%s: House code must be in range [A-P]\n", MyName);
        errno = EINVAL;
        return -1;
    }

    return c;
}

int getdevs(char *list, br_device_list *devs)
{
/*
 * Get a list of devices for an operation to be performed on from
 *  the command line
 */
    char *end;

    if ((br_strtodl(list, devs, &end) < 0)
      || (*end != '\0'))
    {
        fprintf(stderr, "%s: Devices must be in the range of 1-16\n",
          MyName);
        errno = EINVAL;
        return -1;
    }

    return 0;
}

int getdim(char *list, br_device_list *devs, int *dim)
{
/*
 * Get devices that should be dimmed from the command line, and how
 *  much to dim them
 */    
    
    char *end;

    printf("list = %s\n", list);

    *dim = strtol(list, &end, 0);
    printf("end = %s\n", end);

    if (*end != '\0') {
        if (*end == ',') {
            end++;
            if (br_strtodl(end, devs, &end) < 0)
                return -1;
        }
    }
    printf("end = %s\n", end);

    /*
     * May have more dimlevels when I get a chance to play with variable
     *  dimming
     */

    if ((*end != '\0') || (*dim < -DIMRANGE) || (*dim > DIMRANGE))
    {
        fprintf(stderr, "%s: For dimming either specify just a dim level or "
          "a comma\n",MyName);
        fprintf(stderr, "separated list containing the dim level and the "
          "devices to dim.\n");
        fprintf(stderr, "%s: Valid dimlevels are numbers between %d and %d.\n",
          MyName, -DIMRANGE, DIMRANGE);
        errno = EINVAL;
        return -1;
    }

    return 0;
}

int open_port(br_control_info *cinfo, char *port)
{
/*
 * Open the serial port that a FireCracker device is (we expect) on
 */
    int fd;


    if (Verbose >= 2)
        printf("%s: Opening serial port %s.\n", MyName, port);

    /*
     * Oh, yeah.  Don't need O_RDWR for ioctls.  This is safer.
     */
        
    if ((fd = open(port, O_RDONLY | O_NONBLOCK)) < 0) {
        fprintf(stderr, "%s: Error (%s) opening %s.\n", MyName, 
          strerror(errno), port);
        return -1;
    }
    
    /*
     * If we end up with a reserved fd, don't mess with it.  Just to make sure.
     */
    
    if (!SAFE_FILENO(fd)) {
        close(fd);
        errno = EBADF;
        return -1;
    }

    return fd;
}

int close_port(int fd)
{
/*
 * Close the serial port when we're done using it
 */

    if (Verbose >= 2)
        printf("%s: Closing serial port.\n", MyName);

    close(fd);

    return 0;
}

int native_getunits(char *arg, int *house, br_device_list *devs)
{
/*
 * Get units to be accessed from the command line in native BottleRocket style
 */

    if (strlen(arg) < 2) {
        errno = EINVAL;
        return -1;
    }

    if (getdevs(arg + 1, devs) < 0)
        return -1;

    *(arg + 1) = '\0';

    if ((*house = gethouse(arg)) < 0)
        return -1;

    return 0;
}

int native_getcmd(char *arg)
{
/*
 * Convert a native BottleRocket command to the appropriate token
 */

    if (!strcasecmp(arg, "ON"))
        return ON;

    if (!strcasecmp(arg, "OFF"))
        return OFF;

    if (!strcasecmp(arg, "DIM"))
        return DIM;

    if (!strcasecmp(arg, "BRIGHT"))
        return BRIGHT;

    if (!strcasecmp(arg, "ALL_ON"))
        return ALL_ON;

    if (!strcasecmp(arg, "ALL_OFF"))
        return ALL_OFF;

    if (!strcasecmp(arg, "LAMPS_ON"))
        return ALL_LAMPS_ON;

    if (!strcasecmp(arg, "LAMPS_OFF"))
        return ALL_LAMPS_OFF;

    fprintf(stderr, "%s: Native br command must be one of ON, OFF, "
      "DIM, BRIGHT, ALL_ON, ALL_OFF, LAMPS_ON or LAMPS_OFF.\n", MyName);
    errno = EINVAL;

    return -1;
}

int native_cmdline(br_control_info *cinfo, int argc, char *argv[], int optind, br_device_list *devs)
{
/*
 * Get arguments from the command line in native BottleRocket style
 */

    int cmd;
    int house = 0;
    int i;
    
    if (argc - optind < 2) {
        usage();
        errno = EINVAL;
        return -1;
    }

    /*
     * Two arguments at a time; device and command
     */
    
    for (i = optind; i < argc; i += 2) {        
        cmd = native_getcmd(argv[i + 1]);
        
        switch(cmd) {
            case ON:
                if (native_getunits(argv[i], &house, devs) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, ON, house, devs, 0) < 0)
                    return -1;
                break;
        
            case OFF:
                if (native_getunits(argv[i], &house, devs) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, OFF, house, devs, 0) < 0)
                    return -1;
                break;
        
            case DIM:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, DIM, house, NULL, -1) < 0)
                    return -1;
                break;
        
            case BRIGHT:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, DIM, house, NULL, 1) < 0)
                    return -1;
                break;
    
            case ALL_ON:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, ALL_ON, house, NULL, 1) < 0)
                    return -1;
                break;
        
            case ALL_OFF:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, ALL_OFF, house, NULL, 1) < 0)
                    return -1;
                break;
        
            case ALL_LAMPS_ON:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, ALL_LAMPS_ON, house, NULL, 1) < 0)
                    return -1;
                break;
        
            case ALL_LAMPS_OFF:
                if ((house = gethouse(argv[i])) < 0)
                    return -1;
                if (br_add_dl_cmd(cinfo, ALL_LAMPS_OFF, house, NULL, 1) < 0)
                    return -1;
                break;
        
            default:
                errno = EINVAL;
                return -1;
        }
    }

    if (i != argc) {
        errno = EINVAL;
        return -1;
    }

    return 0;
}        

int main(int argc, char **argv)
{
    char *port_source = "at compile time";
    char *tmp_port;
    char *port = X10_PORTNAME;
    int opt;
    int house = 0;
    int repeat;
    int dimlevel = 0;
    int fd;
    br_control_info *cinfo = NULL;
    br_device_list *devs = NULL;
    
#ifdef HAVE_GETOPT_LONG
    int opt_index;
    static struct option long_options[] = {
        {"help",       no_argument,            0, 'h'},
        {"port",       required_argument,      0, 'x'},
        {"repeat",     required_argument,      0, 'r'},
        {"on",         required_argument,      0, 'n'},
        {"off",        required_argument,      0, 'f'},
        {"ON",         no_argument,            0, 'N'},
        {"OFF",        no_argument,            0, 'F'},
        {"dim",        required_argument,      0, 'd'},
        {"lamps_on",   no_argument,            0, 'B'},
        {"lamps_off",  no_argument,            0, 'D'},
        {"inverse",    no_argument,            0, 'i'},
        {"house",      required_argument,      0, 'c'},
        {"verbose",    no_argument,            0, 'v'},
        {0, 0, 0, 0}
    };
#endif

#define OPT_STRING     "x:hvr:ic:n:Nf:Fd:BD"

    /*
     * It's possible to do an exec without even passing argv[0];
     *  this is just to make sure we don't make any bad
     *  output decisions without accounting for that
     */

    if (argc)
        MyName = argv[0];

    if (argc < 2) {
        usage();
        exit(1);
    }

    /*
     * This is where we store all of the info to be passed to the
     *  command engine.
     */

    if ((cinfo = br_new_control_info()) == NULL)
        exit(errno);

    if ((devs = br_new_device_list()) == NULL)
        exit(errno);

    if ((tmp_port = getenv("X10_PORTNAME"))) {
        port_source = "in the environment variable X10_PORTNAME";
        if (!checkimmutableport(port_source))
            port = tmp_port;
    }
    
#ifdef HAVE_GETOPT_LONG
    while ((opt = getopt_long(argc, argv, OPT_STRING, long_options,
      &opt_index)) != -1)
    {
#else
    while ((opt = getopt(argc, argv, OPT_STRING)) != -1) {
#endif        
        switch (opt) {
        case 'x':                                  /* Set the port */
            port_source = "on the command line";
            if (checkimmutableport(port_source) < 0)
                exit(errno);
            port = optarg;
            break;
        case 'r':                                /* Repeat */
            repeat = atoi(optarg);
            if (!repeat && !isdigit(*optarg)) {
                fprintf(stderr, "%s: Invalid repeat value specified.\n",
                  MyName);
                exit(EINVAL);
            }
            cinfo->repeat = (repeat ? repeat:INT_MAX);
            break;
        case 'v':                                /* Verbose */
            if (Verbose++ == 10)
                fprintf(stderr, "\nGet a LIFE already.  "
                  "I've got enough v's, thanks.\n\n");
            break;
        case 'i':
            cinfo->inverse++;    /* no this isn't documented. 
                                  *   your free gift for reading the source.
                                  */
            break;
        case 'c':                                /* Set housecode */
            if ((house = gethouse(optarg)) < 0)
                exit(errno);
            break;
        case 'n':                                /* Device(s) on */
            if (getdevs(optarg, devs) < 0)
                exit(errno);
            if (br_add_dl_cmd(cinfo, ON, house, devs, 0) < 0)
                exit(errno);
            break;
        case 'N':                                /* All on */
            if (br_add_dl_cmd(cinfo, ALL_ON, house, NULL, 0) < 0)
                exit(errno);
            break;
        case 'f':                                /* Device(s) off */
            if (getdevs(optarg, devs) < 0)
                exit(errno);
            if (br_add_dl_cmd(cinfo, OFF, house, devs, 0) < 0)
                exit(errno);
            break;
        case 'F':                                /* All off */
            if (br_add_dl_cmd(cinfo, ALL_OFF, house, NULL, 0) < 0)
                exit(errno);
            break;
        case 'd':                                /* Dim (/bright) */
            if (getdim(optarg, devs, &dimlevel) < 0)
                exit(errno);
            if (add_dimcmd(cinfo, house, devs, dimlevel) < 0)
                exit(errno);
            break;
        case 'B':                                /* All lamps on */
            if (br_add_dl_cmd(cinfo, ALL_LAMPS_ON, house, NULL, 0) < 0)
                exit(errno);
            break;
        case 'D':                                /* All lamps off */
            if (br_add_dl_cmd(cinfo, ALL_LAMPS_OFF, house, NULL, 0) < 0)
                exit(errno);
            break;
        case 'h':                                /* Help */
            usage();
            exit(0);
        default:                                 /* Someone messed up. */
            usage();
            exit(EINVAL);
        }
    }

    if (argc > optind) {
        /*
         * Must be using the native BottleRocket command line...
         */
    
        if (native_cmdline(cinfo, argc, argv, optind, devs) < 0)
            exit(errno);
    }

    if (!br_get_num_devices(devs)) {
        usage();
        exit(0);
    }

    if ((fd = open_port(cinfo, port)) < 0)
        exit(errno);

    if (br_set_fd(cinfo, fd) < 0)
        exit(errno);

    if (br_execute(cinfo) < 0)
        exit(errno);
            
    if (close_port(fd) < 0)
        exit(errno);
    
    br_free_control_info(cinfo);
    br_free_device_list(devs);

    return 0;
}
