/*
 * br_cmd_engine.c -- Command processing functions for BottleRocket.
 *
 * (c) 1999 by Tymm Twillman (tymm@acm.org).  LGPL applies.  No warranties.
 *  TV is bad for your eyes.
 *
 */

#ifdef __cplusplus
extern C {
#endif

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include "br_cmd.h"
#include "br_cmd_engine.h"

int br_inverse_cmd(int cmd)
{
    switch (cmd) {
        case ON:
            return OFF;
            break;
        case OFF:
            return ON;
            break;
        case DIM:
            return BRIGHT;
            break;
        case BRIGHT:
            return DIM;
            break;
        case ALL_ON:
            return ALL_OFF;
            break;
        case ALL_OFF:
            return ALL_ON;
            break;
        case ALL_LAMPS_ON:
            return ALL_LAMPS_OFF;
            break;
        case ALL_LAMPS_OFF:
            return ALL_LAMPS_ON;
            break;
    }

    return -1;
}

int br_set_fd(br_control_info *cinfo, int fd) {
    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_set_fd", "NULL control info pointer");
        return -1;
    }

    cinfo->fd = fd;

    return 0;
}

int br_execute(br_control_info *cinfo)
{
/*
 * Run through a list of commands and execute them
 */

    register int i;
    register int j;
    register int k;
    register int repeat = cinfo->repeat;
    int inverse = cinfo->inverse;
    char unit;
    int rv;

    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_execute", "NULL control info pointer");
        return -1;
    }

    if (cinfo->devs == NULL) {
        errno = EINVAL;
        brerror("br_execute", "NULL device list pointer");
        return -1;
    }

    /* Lotsa loops, woo! */

    /* However many times we have to repeat this thing... */

    for (; repeat > 0; repeat--) {

        /* Do for each command in the command list... */

        for (i = 0; i < cinfo->numcmds; i++) {

            /* For each device in the device list for that command */

            if (cinfo->devs[i] == NULL) {
                errno = EINVAL;
                brerror("br_execute", "NULL device list");
                return -1;
            }

            for (j = 0; j < (CMDHASDEVS(cinfo->cmds[i]) ? cinfo->devs[i]->numdevs:1); j++) {

                unit = ((char)cinfo->houses[i] << 4)
                  | (CMDHASDEVS(cinfo->cmds[i]) ? cinfo->devs[i]->devs[j]:0);

                /* For each dim level if it's a dim command */

                for (k = 0; k < (ISDIMCMD(cinfo->cmds[i]) ? cinfo->dimlevels[i]:1); k++) {
                    rv = x10_br_out(cinfo->fd, unit,
                      (inverse < 0) ? br_inverse_cmd(cinfo->cmds[i]):cinfo->cmds[i]);
                    if (rv < 0)
                        return -1;
                }
            }
        }

        if (inverse) inverse = 0 - inverse;
    }

    return 0;
}

br_device_list *br_new_device_list()
{
    br_device_list *devs;

    devs = malloc(sizeof(br_device_list));

    if (devs == NULL) {
        brerror("br_new_device_list", "malloc");
        return NULL;
    }

    devs->numdevs = 0;
    devs->devs = NULL;

    return devs;
}

int br_free_device_list(br_device_list *devs)
{
    if (devs != NULL) {
        free(devs->devs);
        free(devs);
    }

    return 0;
}

int br_add_device(br_device_list *devs, int dev)
{
    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_add_device", "NULL device list");
        return -1;
    }

    if (devs->numdevs >= devs->allocateddevs) {
        devs->devs = realloc(devs->devs, (devs->numdevs + DEV_BLKSIZE) * sizeof(int));

        if (devs->devs == NULL) {
            brerror("br_add_device", "realloc");
            return -1;
        }
    }

    devs->devs[devs->numdevs] = dev;

    devs->numdevs++;

    return 0;
}

int br_del_device(br_device_list *devs, int dev)
{
    register int i;
    int moveby = 0;


    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_del_device", "NULL device list");
        return -1;
    }

    for (i = 0; i <= (devs->numdevs - moveby); i++) {
        if (devs->devs[i] == dev)
            moveby++;

        devs->devs[i] = devs->devs[i + moveby];
    }

    devs->numdevs -= moveby;

    if (devs->numdevs == 0) {
        free(devs->devs);
        devs->devs = NULL;
        devs->numdevs = 0;
        devs->allocateddevs = 0;
    }

    return 0;
}


br_control_info *br_new_control_info()
{
    br_control_info *cinfo;


    cinfo = malloc(sizeof(br_control_info));

    if (cinfo == NULL) {
        brerror("br_new_control_info", "malloc");
        return NULL;
    }

    cinfo->inverse = 0;
    cinfo->repeat = 1;
    cinfo->fd = 0;
    cinfo->numcmds = 0;
    cinfo->allocatedcmds = 0;
    cinfo->devs = NULL;
    cinfo->houses = NULL;
    cinfo->dimlevels = NULL;
    cinfo->cmds = NULL;

    return cinfo;
}

int br_free_control_info(br_control_info *cinfo)
{
    br_free_cmds(cinfo);
    return 0;
}

int br_malloc_cmds(br_control_info *cinfo, int numcmds)
{
    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_malloc_cmds", "NULL control info pointer");
        return -1;
    }

    cinfo->cmds = malloc(numcmds * sizeof(int));

    if ((cinfo->cmds) == NULL) {
        brerror("br_malloc_cmds", "malloc");
        return -1;
    }

    cinfo->houses = malloc(numcmds * sizeof(int));

    if ((cinfo->houses) == NULL) {
        brerror("br_malloc_cmds", "malloc");
        return -1;
    }

    cinfo->devs = malloc(numcmds * sizeof(br_device_list *));

    if ((cinfo->devs) == NULL) {
        brerror("br_malloc_cmds", "malloc");
        return -1;
    }

    cinfo->dimlevels = malloc(numcmds * sizeof(int));

    if ((cinfo->dimlevels) == NULL) {
        brerror("br_malloc_cmds", "malloc");
        return -1;
    }

    cinfo->allocatedcmds = numcmds;

    return 0;
}

int br_realloc_cmds(br_control_info *cinfo, int numcmds)
{
    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_realloc_cmds", "NULL control info pointer");
        return -1;
    }

    cinfo->cmds = realloc(cinfo->cmds, numcmds * sizeof(int));

    if ((cinfo->cmds) == NULL) {
        brerror("br_realloc_cmds", "realloc");
        return -1;
    }

    cinfo->houses = realloc(cinfo->houses, numcmds * sizeof(int));

    if ((cinfo->houses) == NULL) {
        brerror("br_realloc_cmds", "realloc");
        return -1;
    }

    cinfo->devs = realloc(cinfo->devs, numcmds * sizeof(br_device_list *));

    if ((cinfo->devs) == NULL) {
        brerror("br_realloc_cmds", "realloc");
        return -1;
    }

    cinfo->dimlevels = realloc(cinfo->dimlevels, numcmds * sizeof(int));

    if ((cinfo->dimlevels) == NULL) {
        brerror("br_realloc_cmds", "realloc");
        return -1;
    }

    cinfo->allocatedcmds = numcmds;

    return 0;
}

int br_free_cmds(br_control_info *cinfo)
{
    register int i;


    if (cinfo == NULL) {
        return 0;
    }

    free(cinfo->cmds);
    cinfo->cmds = NULL;
    for (i = 0; i < cinfo->numcmds; i++);
        br_free_device_list(cinfo->devs[i]);
    free(cinfo->devs);
    cinfo->devs = NULL;
    free(cinfo->dimlevels);
    cinfo->dimlevels = NULL;
    free(cinfo->houses);
    cinfo->houses = NULL;

    cinfo->numcmds = 0;
    cinfo->allocatedcmds = 0;

    return 0;
}

int br_add_dl_cmd(br_control_info *cinfo, int cmd, int house, br_device_list *devs, int dimlevel)
{
    /*
     * Add a command, plus devices for it to act on and other info, to the
     *  list of commands to be executed
     */

    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_add_cmd", "NULL control info pointer");
        return -1;
    }

    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_add_cmd", "NULL device list");
        return -1;
    }

    if (cinfo->numcmds >= cinfo->allocatedcmds) {
        if (br_realloc_cmds(cinfo, cinfo->numcmds + CMD_BLKSIZE) < 0)
            return -1;
    }

    cinfo->cmds[cinfo->numcmds] = cmd;
    cinfo->devs[cinfo->numcmds] = devs;
    cinfo->dimlevels[cinfo->numcmds] = dimlevel;
    cinfo->houses[cinfo->numcmds] = house;

    cinfo->numcmds++;

    return 0;
}

int br_add_cmd(br_control_info *cinfo, int cmd, int house, int dev, int dimlevel)
{
    br_device_list *devs;


    devs = br_new_device_list();

    if (devs == NULL)
        return -1;

    br_add_device(devs, dev);

    if (br_add_dl_cmd(cinfo, cmd, house, devs, dimlevel) < 0)
        return -1;

    return 0;
}

int br_del_cmd(br_control_info *cinfo, int index)
{
    register int i;

    if (cinfo == NULL) {
        errno = EINVAL;
        brerror("br_del_cmd", "NULL control info pointer");
        return -1;
    }

    if (index >= cinfo->numcmds) {
        errno = EINVAL;
        brerror("br_del_cmd", "invalid command index");
        return -1;
    }

    if ((cinfo->numcmds - 1) == 0) {
        if (br_free_cmds(cinfo) < 0)
            return -1;

        return 0;
    }

    br_free_device_list(cinfo->devs[index]);

    for (i = index; i < cinfo->numcmds - 1; i++) {
        cinfo->cmds[i] = cinfo->cmds[i + 1];
        cinfo->devs[i] = cinfo->devs[i + 1];
        cinfo->dimlevels[i] = cinfo->dimlevels[i + 1];
        cinfo->houses[i] = cinfo->houses[i + 1];
    }

    cinfo->numcmds--;

    return 0;
}

int br_strtodl(char *dlptr, br_device_list *devs, char **endptr)
{
    int dev;
    char *my_endptr = NULL;


    if (devs == NULL) {
        errno = EINVAL;
        brerror("br_strtodl", "NULL device list");
        return -1;
    }

    /* Get rid of any residue */

    free(devs->devs);
    devs->devs = 0;
    devs->allocateddevs = 0;
    devs->numdevs = 0;

    do {
        dev = strtol(dlptr, &my_endptr, 0);

        if ((dev > 16) || (dev < 1))
        {
            errno = EINVAL;
            brerror("br_strtodl", "Bad device number");
            return -1;
        }

        if ((*my_endptr != '\0') && (*my_endptr != ',')) {
            *endptr = my_endptr;
            return 0;
        }

        if (br_add_device(devs, dev - 1) < 0)
            return -1;

        dlptr = my_endptr;
    } while (*dlptr++);

    *endptr = --dlptr;

    return 0;
}

int br_dlcat(br_device_list *a, br_device_list *b)
{
    register int i;


    if ((a == NULL) || (b == NULL)) {
        errno = EINVAL;
        brerror("br_dlcat", "NULL device list");
        return -1;
    }

    for (i = 0; i < b->numdevs; i++) {
        if (br_add_device(a, b->devs[i]) < 0)
            return -1;
    }

    return 0;
}

int br_strtohc(char *hcptr, char **endptr)
{
    char *my_endptr = hcptr;
    char c;


    while (isspace(*my_endptr))
        my_endptr++;

    if (!*my_endptr) {
        *endptr = hcptr;
        return 0;
    }

    c = toupper(*my_endptr) - 'A';

    if ((c < 0) || (c > 15)) {
        *endptr = hcptr;
        return 0;
    }

    *endptr = ++my_endptr;

    return c;
}

int br_get_next_device(br_device_list *devs, int *index)
{
    if ((devs == NULL) || (index == NULL))
        return -1;

    if (*index >= devs->numdevs)
        return -1;

    return devs->devs[(*index)++];
}

int br_get_num_devices(br_device_list *devs)
{
    if (devs == NULL)
        return 0;

    return devs->numdevs;
}

#ifdef __cplusplus
}
#endif
